@@ -1,9 +1,9 @@
 /*  Lunzip - Decompressor for lzip files
-    Copyright (C) 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2010, 2011, 2012, 2013, 2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -85,6 +85,7 @@ static inline void Rd_load( struct Range_decoder * const rdec )
   for( i = 0; i < 5; ++i )
     rdec->code = (rdec->code << 8) | Rd_get_byte( rdec );
   rdec->range = 0xFFFFFFFFU;
+  rdec->code &= rdec->range;	/* make sure that first byte is discarded */
   }
 
 static inline void Rd_normalize( struct Range_decoder * const rdec )
@@ -140,24 +141,24 @@ static inline int Rd_decode_bit( struct Range_decoder * const rdec,
 static inline int Rd_decode_tree( struct Range_decoder * const rdec,
                                   Bit_model bm[], const int num_bits )
   {
-  int model = 1;
+  int symbol = 1;
   int i;
   for( i = num_bits; i > 0; --i )
-    model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  return model - (1 << num_bits);
+    symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  return symbol - (1 << num_bits);
   }
 
 static inline int Rd_decode_tree6( struct Range_decoder * const rdec,
                                    Bit_model bm[] )
   {
-  int model = 1;
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  model = ( model << 1 ) | Rd_decode_bit( rdec, &bm[model] );
-  return model - (1 << 6);
+  int symbol = 1;
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  symbol = ( symbol << 1 ) | Rd_decode_bit( rdec, &bm[symbol] );
+  return symbol & 0x3F;
   }
 
 static inline int Rd_decode_tree_reversed( struct Range_decoder * const rdec,
@@ -179,9 +180,9 @@ static inline int Rd_decode_tree_reversed4( struct Range_decoder * const rdec,
                                             Bit_model bm[] )
   {
   int model = 1;
-  int symbol = 0;
-  int bit = Rd_decode_bit( rdec, &bm[model] );
-  model = (model << 1) + bit; symbol |= bit;
+  int symbol = Rd_decode_bit( rdec, &bm[model] );
+  int bit;
+  model = (model << 1) + symbol;
   bit = Rd_decode_bit( rdec, &bm[model] );
   model = (model << 1) + bit; symbol |= (bit << 1);
   bit = Rd_decode_bit( rdec, &bm[model] );
@@ -195,8 +196,7 @@ static inline int Rd_decode_matched( struct Range_decoder * const rdec,
   {
   Bit_model * const bm1 = bm + 0x100;
   int symbol = 1;
-  int i;
-  for( i = 7; i >= 0; --i )
+  while( symbol < 0x100 )
     {
     int match_bit, bit;
     match_byte <<= 1;
@@ -210,53 +210,34 @@ static inline int Rd_decode_matched( struct Range_decoder * const rdec,
       break;
       }
     }
-  return symbol - 0x100;
+  return symbol & 0xFF;
   }
 
-
-struct Len_decoder
+static inline int Rd_decode_len( struct Range_decoder * const rdec,
+                                 struct Len_model * const lm,
+                                 const int pos_state )
   {
-  Bit_model choice1;
-  Bit_model choice2;
-  Bit_model bm_low[pos_states][len_low_symbols];
-  Bit_model bm_mid[pos_states][len_mid_symbols];
-  Bit_model bm_high[len_high_symbols];
-  };
-
-static inline void Led_init( struct Len_decoder * const len_decoder )
-  {
-  Bm_init( &len_decoder->choice1 );
-  Bm_init( &len_decoder->choice2 );
-  Bm_array_init( len_decoder->bm_low[0], pos_states * len_low_symbols );
-  Bm_array_init( len_decoder->bm_mid[0], pos_states * len_mid_symbols );
-  Bm_array_init( len_decoder->bm_high, len_high_symbols );
-  }
-
-static inline int Led_decode( struct Len_decoder * const len_decoder,
-                              struct Range_decoder * const rdec,
-                              const int pos_state )
-  {
-  if( Rd_decode_bit( rdec, &len_decoder->choice1 ) == 0 )
-    return Rd_decode_tree( rdec, len_decoder->bm_low[pos_state], len_low_bits );
-  if( Rd_decode_bit( rdec, &len_decoder->choice2 ) == 0 )
+  if( Rd_decode_bit( rdec, &lm->choice1 ) == 0 )
+    return Rd_decode_tree( rdec, lm->bm_low[pos_state], len_low_bits );
+  if( Rd_decode_bit( rdec, &lm->choice2 ) == 0 )
     return len_low_symbols +
-           Rd_decode_tree( rdec, len_decoder->bm_mid[pos_state], len_mid_bits );
+           Rd_decode_tree( rdec, lm->bm_mid[pos_state], len_mid_bits );
   return len_low_symbols + len_mid_symbols +
-         Rd_decode_tree( rdec, len_decoder->bm_high, len_high_bits );
+         Rd_decode_tree( rdec, lm->bm_high, len_high_bits );
   }
 
 
 struct LZ_decoder
   {
   unsigned long long partial_data_pos;
-  int dictionary_size;
+  struct Range_decoder * rdec;
+  unsigned dictionary_size;
   int buffer_size;
   uint8_t * buffer;		/* output buffer */
   int pos;			/* current pos in buffer */
   int stream_pos;		/* first byte not yet written to file */
   uint32_t crc;
   int outfd;			/* output file descriptor */
-  int member_version;
 
   Bit_model bm_literal[1<<literal_context_bits][0x300];
   Bit_model bm_match[states][pos_states];
@@ -265,103 +246,124 @@ struct LZ_decoder
   Bit_model bm_rep1[states];
   Bit_model bm_rep2[states];
   Bit_model bm_len[states][pos_states];
-  Bit_model bm_dis_slot[max_dis_states][1<<dis_slot_bits];
+  Bit_model bm_dis_slot[len_states][1<<dis_slot_bits];
   Bit_model bm_dis[modeled_distances-end_dis_model];
   Bit_model bm_align[dis_align_size];
 
-  struct Range_decoder * range_decoder;
-  struct Len_decoder len_decoder;
-  struct Len_decoder rep_match_len_decoder;
+  struct Len_model match_len_model;
+  struct Len_model rep_len_model;
   };
 
-void LZd_flush_data( struct LZ_decoder * const decoder );
+void LZd_flush_data( struct LZ_decoder * const d );
 
-bool LZd_verify_trailer( struct LZ_decoder * const decoder,
-                         struct Pretty_print * const pp );
+int seek_read( const int fd, uint8_t * const buf, const int size,
+               const int offset );
 
-static inline uint8_t LZd_get_prev_byte( const struct LZ_decoder * const decoder )
+static inline uint8_t LZd_get_prev_byte( const struct LZ_decoder * const d )
   {
-  const int i =
-    ( ( decoder->pos > 0 ) ? decoder->pos : decoder->buffer_size ) - 1;
-  return decoder->buffer[i];
+  const int i = ( ( d->pos > 0 ) ? d->pos : d->buffer_size ) - 1;
+  return d->buffer[i];
   }
 
-static inline uint8_t LZd_get_byte( const struct LZ_decoder * const decoder,
+static inline uint8_t LZd_get_byte( const struct LZ_decoder * const d,
                                     const int distance )
   {
-  int i = decoder->pos - distance - 1;
-  if( i < 0 ) i += decoder->buffer_size;
-  return decoder->buffer[i];
+  uint8_t b;
+  const int i = d->pos - distance - 1;
+  if( i >= 0 ) b = d->buffer[i];
+  else if( i + d->buffer_size >= d->pos )
+    b = d->buffer[i+d->buffer_size];
+  else if( seek_read( d->outfd, &b, 1, i - d->stream_pos ) != 1 )
+    { show_error( "Seek error", errno, false ); cleanup_and_fail( 1 ); }
+  return b;
   }
 
-static inline void LZd_put_byte( struct LZ_decoder * const decoder,
-                                 const uint8_t b )
+static inline void LZd_put_byte( struct LZ_decoder * const d, const uint8_t b )
   {
-  decoder->buffer[decoder->pos] = b;
-  if( ++decoder->pos >= decoder->buffer_size ) LZd_flush_data( decoder );
+  d->buffer[d->pos] = b;
+  if( ++d->pos >= d->buffer_size ) LZd_flush_data( d );
   }
 
-static inline void LZd_copy_block( struct LZ_decoder * const decoder,
+static inline void LZd_copy_block( struct LZ_decoder * const d,
                                    const int distance, int len )
   {
-  int i = decoder->pos - distance - 1;
-  if( i < 0 ) i += decoder->buffer_size;
-  if( len < decoder->buffer_size - max( decoder->pos, i ) &&
-      len <= abs( decoder->pos - i ) )		/* no wrap, no overlap */
+  int i = d->pos - distance - 1;
+  if( i < 0 ) i += d->buffer_size;
+  if( len < d->buffer_size - max( d->pos, i ) && len <= abs( d->pos - i ) )
     {
-    memcpy( decoder->buffer + decoder->pos, decoder->buffer + i, len );
-    decoder->pos += len;
+    memcpy( d->buffer + d->pos, d->buffer + i, len );	/* no wrap, no overlap */
+    d->pos += len;
     }
   else for( ; len > 0; --len )
     {
-    decoder->buffer[decoder->pos] = decoder->buffer[i];
-    if( ++decoder->pos >= decoder->buffer_size ) LZd_flush_data( decoder );
-    if( ++i >= decoder->buffer_size ) i = 0;
+    d->buffer[d->pos] = d->buffer[i];
+    if( ++d->pos >= d->buffer_size ) LZd_flush_data( d );
+    if( ++i >= d->buffer_size ) i = 0;
     }
   }
 
-static inline bool LZd_init( struct LZ_decoder * const decoder,
-                             const File_header header,
-                             struct Range_decoder * const rdec, const int ofd )
+static inline void LZd_copy_block2( struct LZ_decoder * const d,
+                                    const int distance, int len )
   {
-  decoder->partial_data_pos = 0;
-  decoder->dictionary_size = Fh_get_dictionary_size( header );
-  decoder->buffer_size = max( 65536, decoder->dictionary_size );
-  decoder->buffer = (uint8_t *)malloc( decoder->buffer_size );
-  if( !decoder->buffer ) return false;
-  decoder->pos = 0;
-  decoder->stream_pos = 0;
-  decoder->crc = 0xFFFFFFFFU;
-  decoder->outfd = ofd;
-  decoder->member_version = Fh_version( header );
+  if( distance < d->buffer_size )		/* block is in buffer */
+    { LZd_copy_block( d, distance, len ); return; }
+  if( len < d->buffer_size - d->pos )		/* no wrap */
+    {
+    const int offset = d->pos - d->stream_pos - distance - 1;
+    if( len <= -offset )			/* block is in file */
+      {
+      if( seek_read( d->outfd, d->buffer + d->pos, len, offset ) != len )
+        { show_error( "Seek error", errno, false ); cleanup_and_fail( 1 ); }
+      d->pos += len;
+      return;
+      }
+    }
+  for( ; len > 0; --len )
+    LZd_put_byte( d, LZd_get_byte( d, distance ) );
+  }
 
-  Bm_array_init( decoder->bm_literal[0], (1 << literal_context_bits) * 0x300 );
-  Bm_array_init( decoder->bm_match[0], states * pos_states );
-  Bm_array_init( decoder->bm_rep, states );
-  Bm_array_init( decoder->bm_rep0, states );
-  Bm_array_init( decoder->bm_rep1, states );
-  Bm_array_init( decoder->bm_rep2, states );
-  Bm_array_init( decoder->bm_len[0], states * pos_states );
-  Bm_array_init( decoder->bm_dis_slot[0], max_dis_states * (1 << dis_slot_bits) );
-  Bm_array_init( decoder->bm_dis, modeled_distances - end_dis_model );
-  Bm_array_init( decoder->bm_align, dis_align_size );
+static inline bool LZd_init( struct LZ_decoder * const d,
+                             struct Range_decoder * const rde,
+                             const int buffer_size,
+                             const int dict_size, const int ofd )
+  {
+  d->partial_data_pos = 0;
+  d->rdec = rde;
+  d->dictionary_size = dict_size;
+  d->buffer_size = min( buffer_size, max( 65536, dict_size ) );
+  d->buffer = (uint8_t *)malloc( d->buffer_size );
+  if( !d->buffer ) return false;
+  d->pos = 0;
+  d->stream_pos = 0;
+  d->crc = 0xFFFFFFFFU;
+  d->outfd = ofd;
 
-  decoder->range_decoder = rdec;
-  Led_init( &decoder->len_decoder );
-  Led_init( &decoder->rep_match_len_decoder );
-  decoder->buffer[decoder->buffer_size-1] = 0;	/* prev_byte of first_byte */
+  Bm_array_init( d->bm_literal[0], (1 << literal_context_bits) * 0x300 );
+  Bm_array_init( d->bm_match[0], states * pos_states );
+  Bm_array_init( d->bm_rep, states );
+  Bm_array_init( d->bm_rep0, states );
+  Bm_array_init( d->bm_rep1, states );
+  Bm_array_init( d->bm_rep2, states );
+  Bm_array_init( d->bm_len[0], states * pos_states );
+  Bm_array_init( d->bm_dis_slot[0], len_states * (1 << dis_slot_bits) );
+  Bm_array_init( d->bm_dis, modeled_distances - end_dis_model );
+  Bm_array_init( d->bm_align, dis_align_size );
+
+  Lm_init( &d->match_len_model );
+  Lm_init( &d->rep_len_model );
+  d->buffer[d->buffer_size-1] = 0;		/* prev_byte of first byte */
   return true;
   }
 
-static inline void LZd_free( struct LZ_decoder * const decoder )
-  { free( decoder->buffer ); }
+static inline void LZd_free( struct LZ_decoder * const d )
+  { free( d->buffer ); }
 
-static inline unsigned LZd_crc( const struct LZ_decoder * const decoder )
-  { return decoder->crc ^ 0xFFFFFFFFU; }
+static inline unsigned LZd_crc( const struct LZ_decoder * const d )
+  { return d->crc ^ 0xFFFFFFFFU; }
 
 static inline unsigned long long
-LZd_data_position( const struct LZ_decoder * const decoder )
-  { return decoder->partial_data_pos + decoder->pos; }
+LZd_data_position( const struct LZ_decoder * const d )
+  { return d->partial_data_pos + d->pos; }
 
-int LZd_decode_member( struct LZ_decoder * const decoder,
+int LZd_decode_member( struct LZ_decoder * const d,
                        struct Pretty_print * const pp );
