@@ -1,9 +1,9 @@
 /*  Lunzip - Decompressor for lzip files
-    Copyright (C) 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2010, 2011, 2012, 2013, 2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -15,7 +15,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused lunzip to panic.
@@ -51,9 +51,13 @@
 #endif
 
 #include "carg_parser.h"
-#include "lunzip.h"
+#include "lzip.h"
 #include "decoder.h"
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
@@ -61,15 +65,9 @@
 
 const char * const Program_name = "Lunzip";
 const char * const program_name = "lunzip";
-const char * const program_year = "2013";
+const char * const program_year = "2014";
 const char * invocation_name = 0;
 
-#ifdef O_BINARY
-const int o_binary = O_BINARY;
-#else
-const int o_binary = 0;
-#endif
-
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
@@ -86,7 +84,18 @@ bool delete_output_on_interrupt = false;
 
 static void show_help( void )
   {
-  printf( "%s - Decompressor for lzip files.\n", Program_name );
+  printf( "Lunzip is a decompressor for lzip files. It is written in C and its\n"
+          "small size makes it well suited for embedded devices or software\n"
+          "installers that need to decompress files but do not need compression\n"
+          "capabilities. Lunzip is fully compatible with lzip-1.4 or newer.\n"
+          "\nLunzip provides a 'low memory' mode able to decompress any file using as\n"
+          "little memory as 50 kB, irrespective of the dictionary size used to\n"
+          "compress the file. To activate it, specify the size of the output buffer\n"
+          "with the '--buffer-size' option and lunzip will use the decompressed\n"
+          "file as dictionary for distances beyond the buffer size. Of course, the\n"
+          "smaller the output buffer size used in relation to the dictionary size,\n"
+          "the more accesses to disk are needed and the slower the decompression is.\n"
+          "This 'low memory' mode only works when decompressing to a regular file.\n" );
   printf( "\nUsage: %s [options] [files]\n", invocation_name );
   printf( "\nOptions:\n"
           "  -h, --help                 display this help and exit\n"
@@ -98,9 +107,16 @@ static void show_help( void )
           "  -o, --output=<file>        if reading stdin, place the output into <file>\n"
           "  -q, --quiet                suppress all messages\n"
           "  -t, --test                 test compressed file integrity\n"
+          "  -u, --buffer-size=<bytes>  set output buffer size in bytes\n"
           "  -v, --verbose              be verbose (a 2nd -v gives more)\n"
           "If no file names are given, lunzip decompresses from standard input to\n"
           "standard output.\n"
+          "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
+          "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
+          "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+          "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+          "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+          "caused lunzip to panic.\n"
           "\nReport bugs to lzip-bug@nongnu.org\n"
           "Lunzip home page: http://www.nongnu.org/lzip/lunzip.html\n" );
   }
@@ -108,29 +124,95 @@ static void show_help( void )
 
 static void show_version( void )
   {
-  printf( "%s %s\n", Program_name, PROGVERSION );
+  printf( "%s %s\n", program_name, PROGVERSION );
   printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
           "This is free software: you are free to change and redistribute it.\n"
           "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-void show_header( const File_header header )
+static void show_header( const unsigned dictionary_size )
   {
   const char * const prefix[8] =
     { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
   enum { factor = 1024 };
   const char * p = "";
   const char * np = "  ";
-  unsigned num = Fh_get_dictionary_size( header ), i;
+  unsigned num = dictionary_size, i;
   bool exact = ( num % factor == 0 );
 
   for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
     { num /= factor; if( num % factor != 0 ) exact = false;
       p = prefix[i]; np = ""; }
-  fprintf( stderr, "version %d, dictionary size %s%4u %sB.  ",
-           Fh_version( header ), np, num, p );
+  fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
+  }
+
+
+static unsigned long getnum( const char * const ptr,
+                             const unsigned long llimit,
+                             const unsigned long ulimit )
+  {
+  unsigned long result;
+  char * tail;
+  errno = 0;
+  result = strtoul( ptr, &tail, 0 );
+  if( tail == ptr )
+    {
+    show_error( "Bad or missing numerical argument.", 0, true );
+    exit( 1 );
+    }
+
+  if( !errno && tail[0] )
+    {
+    int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
+    int exponent = 0, i;
+    bool bad_multiplier = false;
+    switch( tail[0] )
+      {
+      case ' ': break;
+      case 'Y': exponent = 8; break;
+      case 'Z': exponent = 7; break;
+      case 'E': exponent = 6; break;
+      case 'P': exponent = 5; break;
+      case 'T': exponent = 4; break;
+      case 'G': exponent = 3; break;
+      case 'M': exponent = 2; break;
+      case 'K': if( factor == 1024 ) exponent = 1; else bad_multiplier = true;
+                break;
+      case 'k': if( factor == 1000 ) exponent = 1; else bad_multiplier = true;
+                break;
+      default : bad_multiplier = true;
+      }
+    if( bad_multiplier )
+      {
+      show_error( "Bad multiplier in numerical argument.", 0, true );
+      exit( 1 );
+      }
+    for( i = 0; i < exponent; ++i )
+      {
+      if( ulimit / factor >= result ) result *= factor;
+      else { errno = ERANGE; break; }
+      }
+    }
+  if( !errno && ( result < llimit || result > ulimit ) ) errno = ERANGE;
+  if( errno )
+    {
+    show_error( "Numerical argument out of limits.", 0, false );
+    exit( 1 );
+    }
+  return result;
+  }
+
+
+static int get_dict_size( const char * const arg )
+  {
+  char * tail;
+  int bits = strtol( arg, &tail, 0 );
+  if( bits >= min_dictionary_bits &&
+      bits <= max_dictionary_bits && *tail == 0 )
+    return ( 1 << bits );
+  return getnum( arg, min_dictionary_size, max_dictionary_size );
   }
 
 
@@ -148,11 +230,10 @@ static int extension_index( const char * const name )
   }
 
 
-static int open_instream( const char * const name,
-                          struct stat * const in_statsp, const bool to_stdout )
+static int open_instream( const char * const name, struct stat * const in_statsp,
+                          const bool no_ofile )
   {
-  int infd = -1;
-  infd = open( name, O_RDONLY | o_binary );
+  int infd = open( name, O_RDONLY | O_BINARY );
   if( infd < 0 )
     {
     if( verbosity >= 0 )
@@ -166,12 +247,12 @@ static int open_instream( const char * const name,
     const bool can_read = ( i == 0 &&
                             ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                               S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-    if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
+    if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
       {
       if( verbosity >= 0 )
         fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
                  program_name, name,
-                 ( can_read && !to_stdout ) ?
+                 ( can_read && !no_ofile ) ?
                  " and '--stdout' was not specified" : "" );
       close( infd );
       infd = -1;
@@ -221,7 +302,7 @@ static void set_d_outname( const char * const name, const int i )
 
 static bool open_outstream( const bool force )
   {
-  int flags = O_CREAT | O_WRONLY | o_binary;
+  int flags = O_APPEND | O_CREAT | O_RDWR | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename, flags, outfd_mode );
@@ -260,10 +341,14 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
   bool warning = false;
   if( in_statsp )
     {
+    const mode_t mode = in_statsp->st_mode;
     /* fchown will in many cases return with EPERM, which can be safely ignored. */
-    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
-          errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
+    if( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) == 0 )
+      { if( fchmod( outfd, mode ) != 0 ) warning = true; }
+    else
+      if( errno != EPERM ||
+          fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
+        warning = true;
     }
   if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
   outfd = -1;
@@ -280,8 +365,8 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
   }
 
 
-static int decompress( const int infd, struct Pretty_print * const pp,
-                       const bool testing )
+static int decompress( const int buffer_size, const int infd,
+                       struct Pretty_print * const pp, const bool testing )
   {
   unsigned long long partial_file_pos = 0;
   struct Range_decoder rdec;
@@ -289,13 +374,14 @@ static int decompress( const int infd, struct Pretty_print * const pp,
   bool first_member;
   if( !Rd_init( &rdec, infd ) )
     {
-    show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+    show_error( "Not enough memory.", 0, false );
     cleanup_and_fail( 1 );
     }
 
   for( first_member = true; ; first_member = false )
     {
     int result;
+    unsigned dictionary_size;
     File_header header;
     struct LZ_decoder decoder;
     Rd_reset_member_position( &rdec );
@@ -303,13 +389,14 @@ static int decompress( const int infd, struct Pretty_print * const pp,
     if( Rd_finished( &rdec ) )			/* End Of File */
       {
       if( first_member )
-        { Pp_show_msg( pp, "Error reading member header" ); retval = 1; }
+        { Pp_show_msg( pp, "File ends unexpectedly at member header." );
+          retval = 2; }
       break;
       }
     if( !Fh_verify_magic( header ) )
       {
       if( !first_member ) break;		/* trailing garbage */
-      Pp_show_msg( pp, "Bad magic number (file not in lzip format)" );
+      Pp_show_msg( pp, "Bad magic number (file not in lzip format)." );
       retval = 2; break;
       }
     if( !Fh_verify_version( header ) )
@@ -320,17 +407,19 @@ static int decompress( const int infd, struct Pretty_print * const pp,
                    Fh_version( header ) ); }
       retval = 2; break;
       }
-    if( Fh_get_dictionary_size( header ) < min_dictionary_size ||
-        Fh_get_dictionary_size( header ) > max_dictionary_size )
-      { Pp_show_msg( pp, "Invalid dictionary size in member header" );
+    dictionary_size = Fh_get_dictionary_size( header );
+    if( dictionary_size < min_dictionary_size ||
+        dictionary_size > max_dictionary_size )
+      { Pp_show_msg( pp, "Invalid dictionary size in member header." );
         retval = 2; break; }
 
     if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
-      { Pp_show_msg( pp, 0 ); if( verbosity >= 2 ) show_header( header ); }
+      { Pp_show_msg( pp, 0 );
+        if( verbosity >= 3 ) show_header( dictionary_size ); }
 
-    if( !LZd_init( &decoder, header, &rdec, outfd ) )
+    if( !LZd_init( &decoder, &rdec, buffer_size, dictionary_size, outfd ) )
       {
-      show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+      show_error( "Not enough memory. Try a smaller output buffer size.", 0, false );
       cleanup_and_fail( 1 );
       }
     result = LZd_decode_member( &decoder, pp );
@@ -342,21 +431,19 @@ static int decompress( const int infd, struct Pretty_print * const pp,
         {
         Pp_show_msg( pp, 0 );
         if( result == 2 )
-          fprintf( stderr, "File ends unexpectedly at pos %llu\n",
+          fprintf( stderr, "File ends unexpectedly at pos %llu.\n",
                    partial_file_pos );
         else
-          fprintf( stderr, "Decoder error at pos %llu\n", partial_file_pos );
+          fprintf( stderr, "Decoder error at pos %llu.\n", partial_file_pos );
         }
       retval = 2; break;
       }
     if( verbosity >= 2 )
-      { if( testing ) fprintf( stderr, "ok\n" );
-        else fprintf( stderr, "done\n" ); Pp_reset( pp ); }
+      { fprintf( stderr, testing ? "ok\n" : "done\n" ); Pp_reset( pp ); }
     }
   Rd_free( &rdec );
   if( verbosity == 1 && retval == 0 )
-    { if( testing ) fprintf( stderr, "ok\n" );
-      else fprintf( stderr, "done\n" ); }
+    fprintf( stderr, testing ? "ok\n" : "done\n" );
   return retval;
   }
 
@@ -377,28 +464,6 @@ static void set_signals( void )
   }
 
 
-void Pp_init( struct Pretty_print * const pp, const char * const filenames[],
-              const int num_filenames, const int v )
-  {
-  unsigned stdin_name_len;
-  int i;
-  pp->name = 0;
-  pp->stdin_name = "(stdin)";
-  pp->longest_name = 0;
-  pp->verbosity = v;
-  pp->first_post = false;
-  stdin_name_len = strlen( pp->stdin_name );
-
-  for( i = 0; i < num_filenames; ++i )
-    {
-    const char * const s = filenames[i];
-    const int len = ( (strcmp( s, "-" ) == 0) ? stdin_name_len : strlen( s ) );
-    if( len > pp->longest_name ) pp->longest_name = len;
-    }
-  if( pp->longest_name == 0 ) pp->longest_name = stdin_name_len;
-  }
-
-
 void show_error( const char * const msg, const int errcode, const bool help )
   {
   if( verbosity >= 0 )
@@ -406,7 +471,7 @@ void show_error( const char * const msg, const int errcode, const bool help )
     if( msg && msg[0] )
       {
       fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+      if( errcode > 0 ) fprintf( stderr, ": %s.", strerror( errcode ) );
       fprintf( stderr, "\n" );
       }
     if( help )
@@ -419,7 +484,7 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void internal_error( const char * const msg )
   {
   if( verbosity >= 0 )
-    fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
+    fprintf( stderr, "%s: internal error: %s\n", program_name, msg );
   exit( 3 );
   }
 
@@ -430,6 +495,7 @@ int main( const int argc, const char * const argv[] )
   const char * default_output_filename = "";
   const char ** filenames = 0;
   int num_filenames = 0;
+  int buffer_size = max_dictionary_size;
   int infd = -1;
   int argind = 0;
   int retval = 0;
@@ -448,9 +514,11 @@ int main( const int argc, const char * const argv[] )
     { 'f', "force",           ap_no  },
     { 'h', "help",            ap_no  },
     { 'k', "keep",            ap_no  },
+    { 'n', "threads",         ap_yes },
     { 'o', "output",          ap_yes },
     { 'q', "quiet",           ap_no  },
     { 't', "test",            ap_no  },
+    { 'u', "buffer-size",     ap_yes },
     { 'v', "verbose",         ap_no  },
     { 'V', "version",         ap_no  },
     {  0 ,  0,                ap_no  } };
@@ -461,7 +529,7 @@ int main( const int argc, const char * const argv[] )
   CRC32_init();
 
   if( !ap_init( &parser, argc, argv, options, 0 ) )
-    { show_error( "Memory exhausted.", 0, false ); return 1; }
+    { show_error( "Not enough memory.", 0, false ); return 1; }
   if( ap_error( &parser ) )				/* bad option */
     { show_error( ap_error( &parser ), 0, true ); return 1; }
 
@@ -477,12 +545,14 @@ int main( const int argc, const char * const argv[] )
       case 'f': force = true; break;
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
+      case 'n': break;
       case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
       case 't': testing = true; break;
+      case 'u': buffer_size = get_dict_size( arg ); break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      default : internal_error( "uncaught option" );
+      default : internal_error( "uncaught option." );
       }
     } /* end process options */
 
@@ -504,11 +574,24 @@ int main( const int argc, const char * const argv[] )
     if( strcmp( filenames[i], "-" ) != 0 ) filenames_given = true;
     }
 
+  if( buffer_size < max_dictionary_size )
+    {
+    if( to_stdout || testing )
+      { show_error( "'--buffer-size' is incompatible with '--stdout' and '--test'.", 0, false );
+        return 1; }
+    if( !default_output_filename[0] )
+      for( i = 0; i < num_filenames; ++i )
+        if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
+          { show_error( "Output file must be specified when decompressing from stdin with a\n"
+                        "        reduced buffer size.", 0, false );
+            return 1; }
+    }
+
   if( !to_stdout && !testing &&
       ( filenames_given || default_output_filename[0] ) )
     set_signals();
 
-  Pp_init( &pp, filenames, num_filenames, verbosity );
+  Pp_init( &pp, filenames, num_filenames );
 
   output_filename = resize_buffer( output_filename, 1 );
   for( i = 0; i < num_filenames; ++i )
@@ -544,7 +627,7 @@ int main( const int argc, const char * const argv[] )
     else
       {
       input_filename = filenames[i];
-      infd = open_instream( input_filename, &in_stats, to_stdout );
+      infd = open_instream( input_filename, &in_stats, to_stdout || testing );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
       if( !testing )
         {
@@ -573,7 +656,7 @@ int main( const int argc, const char * const argv[] )
       delete_output_on_interrupt = true;
     in_statsp = input_filename[0] ? &in_stats : 0;
     Pp_set_name( &pp, input_filename );
-    tmp = decompress( infd, &pp, testing );
+    tmp = decompress( buffer_size, infd, &pp, testing );
     if( tmp > retval ) retval = tmp;
     if( tmp && !testing ) cleanup_and_fail( retval );
 
